//
// star strike 2000
// by Kevin Dixon
// for Allegro SizeHack 2000
//

#include <cmath>
#include <ctime>
#include <stl.h>
#include "allegro.h"

#define W 640
#define H 480

#define HS  40
#define ES 100
#define EF  80

#define TC(a,b,c,d) textout_centre(a, font, b, W/2, c, d)

typedef BITMAP* BM;

struct Sh;
struct Bs;
struct St;
struct Ob;

// graphics data file
DATAFILE *dt;

// fb: frame buffer for double buffering 
// bg: bitmap for generator
// ex: bitmap for explosions
BM fb, bg, ex[EF];

// interrupt counters for game timer and FPS
volatile int tmr;

// list of game objects
vector<Ob*> bl;

// level information
// wv: current wave
// dim: dimensions of the universe
// gn: number of generators remaining to be destroyed
// li: player lives
// sc: player score
// died: player has died count
// fin: finished wave count
int wv, dim, gn, li, sc, died, fin;

Sh* sh; // Ship
Bs* bs; // Base
vector<St*>st; // Stars

// Point
struct Pt {
  int x, y, dx, dy;
  Pt(int m=0, int n=0) : x(m), y(n) {}
};

// Distance check for collision detection
int
Di(int x1, int y1, int x2, int y2)
{
  return (int)hypot(x1-x2, y1-y2);
}

// Game Object
struct Ob {
  // mx, my: map sector position
  // x, y: screen position
  // dx, dy: velocity
  // dd: dead flag
  // hp: hit points
  // ra: object radius
  // a: angle
  // t: generic timer
  // h: health bar timer
  int mx, my, x, y, dx, dy, dd, hp, ra, a, t, h;

  Ob(int k, int l, 
     int m=(rand()%W)/2+W/4, int n=(rand()%H)/2+H/4, 
     int p=0, int q=0) 
    : mx(k), my(l), x(m), y(n), dx(p), dy(q), dd(0), h(0) {}

  virtual int Mv() { return 0; }
  virtual void Dr() = 0;

  // bounds check
  void Bound()
  {
    if (y<0) {
      if (my==0) {
        dy = -dy;
        y=0;
      }
      else {
        y=(H-16)<<16;
        my--;
      }
    }
    else if (y > (H-16)<<16) {
      if (my==dim-1) {
        dy = -dy;
        y=(H-16)<<16;
      }
      else {
        y=0;
        my++;
      }
    }

    if (x<0) {
      if (mx==0) {
        dx = -dx;
        x=0;
      }
      else {
        x=(W-16)<<16;
        mx--;
      }
    }
    else if (x > (W-16)<<16) {
      if (mx==dim-1) {
        dx = -dx;
        x=(W-16)<<16;
      }
     else {
        x=0;
        mx++;
      }
    }

    x+=dx;
    y+=dy;
  }

  // Gravity
  void Gr(int m, int n)
  {
    if (x>>16 > m) dx -= 1<<9;
    else dx += 1<<9;
    	
    if (y>>16 > n) dy -= 1<<9;
    else dy += 1<<9;
  }

  void Hit(int k)
  {
    hp-=k;
    if (hp<=0) {
      dd=1;
      hp=0;
    }

    h+=90;
  }

  virtual void Chk();

  void Boom(int, int, int, int);

  // Draw health bar
  void Hb(int m)
  {
    rectfill(fb, x-ra+8, y-ra, x-ra+(hp*ra*2/m)+8, y-ra+3, 11);
    rect(fb, x-ra+7, y-ra, x+ra+9, y-ra+4, 2);
  }
};

// explosion
struct Ex : Ob {
  Ex(int k, int l, int m, int n, int p, int q) : Ob(k, l, m, n, p, q)
  {
    a = rand()%256;
    t = 0;    
  }
  
  int Mv()
  {
    x+=dx;
    y+=dy;

    dx*=.99;
    dy*=.99;

    if(++t >= EF) dd=1;

    return dd;
  }

  void Dr()
  {
    rotate_sprite(fb, ex[t], (x>>16)-ES/2, (y>>16)-ES/2, a<<16);
  }

};

// player ship bullet
struct Fire : Ob {
  Fire(int k, int l, int m, int n, int p, int q) : Ob(k, l, m, n, p, q) {}

  int Mv()
  {
    x+=dx;
    y+=dy;

    if (y < 0      || 
        y > H<<16 || 
        x < 0      || 
        x > W<<16 ) dd=1;

    Chk();

    return dd;
  }

  void Dr()
  {
    draw_sprite(fb, (BM)dt[1].dat, x>>16, y>>16);
  }

  void Chk();
};

// player ship
struct Sh : Ob {
  Sh(int m, int n) : Ob(0, 0, m, n)
  {
    a = 0;
    t = 0;
    hp = 2500;
    died = 0;
  }

  int Mv()
  {
    t++;
    t&=15;

    if(dd && !died) {
      died=200;
      Boom(x, y, dx, dy);
    }

    if(dd) {
      died--;
      if(died == 1) return 1;
    }
    else {
      
      if(!t && key[KEY_SPACE]) {
	bl.push_back(new Fire(mx, my, x+(6<<16), y+(7<<16), 
			      fsin(a<<16)*6, -fcos(a<<16)*6));
	Hit(1);
      }

      if(key[KEY_UP]) {
	dx += ( fsin(a<<16)) >> 4;
	dy += (-fcos(a<<16)) >> 4;
	if(!t) Hit(1);
      }

      if(key[KEY_DOWN]) {
	dx -= ( fsin(a<<16)) >> 5;
	dy -= (-fcos(a<<16)) >> 5;
	if(!t) Hit(1);
      }

      if(!(t&3) && key[KEY_LEFT]) {
	a-=8;
      }

      if(!(t&3) && key[KEY_RIGHT]) {
	a+=8;
      }
      
      if(key[KEY_ENTER]) {
	Hit(100);
	if (!dd) {
	  int ox = mx;
	  text_mode(0);
	  tmr=-20;
	  while(tmr<0) {
	    mx = rand()%dim;
	    my = rand()%dim;
	    textprintf(screen, font, 616, 0, 6, "%i%i", mx, my);
	  }
	  if (ox==mx) mx = (ox+1)%dim;
	  text_mode(-1);
	}
      }
      
      Bound();

      dx*=.995;
      dy*=.995;

      a&=0xff;
    }

    return 0;
  }

  void Dr()
  {
    if (!dd) rotate_sprite(fb, (BM)dt[5].dat, x>>16, y>>16, a<<16);
  }

};

// background stars
struct St : Ob {
  // s: list of star screen positions
  vector<Pt> s;

  St(int n) : Ob(0, 0)
  {
    for(int i=0; i<n; i++) {
      s.push_back(Pt(rand()%W, rand()%H));
    }
  }
  
  void Dr()
  {
    for(int i=0; i<s.size(); i++) putpixel(fb, s[i].x, s[i].y, 9);
  }
  
};

// enemy drone
struct Drone : Ob {
  // lf: firing laser flag
  // lc: counter for duration of laser shot
  // lo: laser overheat flag
  int lf, lc, lo;
  
  Drone(int k, int l, int m, int n) : Ob(k, l, m, n) 
  {
    lf = lc = lo = 0;
    hp = 2;

    dx = (rand()%3-1)<<17;
    dy = (rand()%3-1)<<17;
  }

  int Mv()
  {
    int z = 1000; // acceleration rate

    if (lc>0 && !lf) if (lc-- == 1) lo=0;
    if (sh->mx == mx && sh->my == my) {
      if (!lo && wv>1 && hp>1 && Di(x,y,sh->x,sh->y)>>16 < 100) {
	lf=1;
	sh->Hit(1);
	if (lc++>100) lo=1;
      }
      else lf=0;
    

      if (sh->x > x) dx += z;
      else dx -= z;
	
      if (sh->y > y) dy += z;
      else dy -= z;

      // collision check with player ship
      if (Di(x,y,sh->x,sh->y)>>16 < 12 && !died) {
	dd=1;
	sh->Hit(500);
	Boom(x, y, dx, dy);
      }

    }
    else {
      lf=0;

      if (sh->mx > mx) dx += z;
      else if (sh->mx < mx) dx -= z;
      
      if (sh->my > my) dy += z;
      else if (sh->my < my) dy -= z;
    }

    Bound();

    dx*=.997;
    dy*=.997;

    return dd;
  }

  void Dr()
  {
    draw_sprite(fb, (BM)dt[6].dat, x>>16, y>>16);
    if (lf==1 && !died) line(fb, 
			     (x>>16)+8, (y>>16)+8, 
			     (sh->x>>16)+8, (sh->y>>16)+8, 11);
    
  }

};

// player base
struct Bs : Ob {
  Bs() : Ob(0, 0) {}

  int Mv()
  {
    if (sh->mx==mx && sh->my==my) {
      if (sh->x>>16 > x-4 &&
	  sh->x>>16 < x+4 &&
	  sh->y>>16 > y-4 &&
	  sh->y>>16 < y+4) {
	sh->dx *= 0.9;
	sh->dy *= 0.9;
	if(sh->hp<2500) sh->hp++;
      }
      else {
	sh->Gr(x, y);
      }
    }

    for(int i=0; i<bl.size(); i++) {
      Ob* b = bl[i];
      if(b->mx == mx && b->my == my) {
	if (dynamic_cast<Drone*>(b)) {
	  if (Di((b->x>>16)-6, b->y>>16, x, y) <21) {
	    b->dd=1;
	    Boom(b->x, b->y, b->dx, b->dy);
	  }
	}
      }
    }

    return 0;
  }

  void Dr()
  {
    stretch_sprite(fb, (BM)dt[0].dat, x-40, y-24, 64, 64);
  }

};

// enemy generator
struct Gen : Ob {
  Gen(int k, int l) : Ob(k, l) 
  {
    a = 0;
    hp = 25;
    ra = 14;
  }

  int Mv()
  {
    if (h) h--;

    if (sh->mx==mx && sh->my==my && !died) {
      sh->Gr(x, y);
      Chk();
    }

    a++;
    a&=511;

    if (dd) {
      gn--;
      sc+=200;
    }
    return dd;
  }

  void Dr()
  {
    rotate_sprite(fb, bg, x-8, y-8, a<<15);
    if (h) Hb(25);
  }

};

// asteroid
struct Rock : Ob {
  Rock(int k, int l) : Ob(k, l) 
  {
    a = (rand()%256)<<16;
    hp = 500;
    ra = 28;
  }

  int Mv()
  {
    if (h) h--;

    if (sh->mx==mx && sh->my==my && !died) {
      sh->Gr(x, y);
      Chk();
    }

    if (dd) sc+=100;
    return dd;
  }

  void Dr()
  {
    rotate_scaled_sprite(fb, (BM)dt[4].dat, x-24, y-24, a, 2<<16);
    if (h) Hb(500);
  }

};

// enemy base
struct En : Ob {
  En(int k, int l) : Ob(k, l)
  {
    hp = 50 + wv*5;
    ra = 22;
    t  = 2000/wv;
  }

  int Mv()
  {
    if (h) h--;

    if (t-- == 0) {
      bl.push_back(new Drone(mx, my, x<<16, y<<16));
      t = 4000/wv;
    }

    if (sh->mx==mx && sh->my==my && !gn) {
      sh->Gr(x, y);
    
      if (Di(sh->x>>16, sh->y>>16, x, y) <ra && !died) {
	sh->dd=1;
	Boom(sh->x, sh->y, sh->dx, sh->dy);
      }
    }

    if (dd) {
      fin = 300;
      sc+=500;
    }
    return dd;
  }

  void Dr()
  {
    if (!gn) {
      stretch_sprite(      fb, (BM)dt[2].dat, x-8,  y-8, 32, 32);
      draw_sprite(         fb, (BM)dt[3].dat, x-16, y-16);
      draw_sprite_v_flip(  fb, (BM)dt[3].dat, x-16, y+16);
      draw_sprite_h_flip(  fb, (BM)dt[3].dat, x+16, y-16);
      draw_sprite_vh_flip( fb, (BM)dt[3].dat, x+16, y+16);    
    }
    if (h) Hb(50+wv*5);
  }

};

// check for rock/gen to ship/drone collision
void
Ob::Chk()
{
  for(int i=0; i<bl.size(); i++) {
    Ob* b = bl[i];
    if(b->mx == mx && b->my == my) {
      if (dynamic_cast<Sh*>(b) || 
	  dynamic_cast<Drone*>(b)) {
	if (Di(b->x>>16, b->y>>16, x, y) <ra) {
	  b->dd=1;
	  Boom(b->x, b->y, b->dx, b->dy);
	}
      }
    }
  }
}

// add an explosion to object list
void 
Ob::Boom(int m, int n, int p, int q)
{
  bl.push_back(new Ex(mx, my, m, n, p, q));
}

// check for fire to rock/gen/enemy/drone collision
void
Fire::Chk()
{
  for(int i=0; i<bl.size(); i++) {
    Ob* b = bl[i];
    if(b->mx == mx && b->my == my) {
      Drone* d = dynamic_cast<Drone*>(b);
      if (d) {
	if (Di(d->x, d->y, x, y) >>16 < 8) {
	  d->Hit(1);
	  dd=1;
	  Boom(x, y, d->dx, d->dy);
	  sc+=5;
	}
      }
      
      Gen* g = dynamic_cast<Gen*>(b);
      Rock* r = dynamic_cast<Rock*>(b);
      En* e = dynamic_cast<En*>(b);
    
      if (g || r || (e && !gn)) {
	if (Di(b->x, b->y, (x>>16)-7, (y>>16)-7) < b->ra) {
	  b->Hit(1);
	  dd = 1;
	  Boom(x, y, 0, 0);
	  sc++;
	}
      }      
    }
  }

}

// create player ship
void
SCreate()
{
  sh = new Sh(bs->x<<16, bs->y<<16);
  bl.push_back(sh);
}

// create level for wave n
void
LCreate(int n)
{
  wv = n;
  dim = wv+1;
  if (dim>8) dim=8;
  gn = dim-1;
  fin = 0;
  vector<Pt> s;
  
  for (int x=dim-1; x>=0; x--) {
    for (int y=dim-1; y>=0; y--) {
      s.push_back(Pt(x,y));
      st.push_back(new St(100));
    }
  }
  
  s.pop_back(); // get rid of point 0,0 because thats where the base goes
  random_shuffle(s.begin(), s.end());
  
  bs = new Bs();
  bl.push_back(bs);
  bl.push_back(new En(s[0].x, s[0].y));
  
  for (int i=1; i<gn+1; i++)
    bl.push_back(new Gen(s[i].x, s[i].y));
  
  for (int i=gn+1; i<s.size(); i++) 
    if (rand()%2) bl.push_back(new Rock(s[i].x, s[i].y));
  
  SCreate();
}

// clean up current level
void
LDestroy() 
{
  while(bl.size() !=0) {
    delete *bl.begin();
    bl.erase(bl.begin());
  }

  while (st.size() != 0) {
    delete *st.begin();
    st.erase(st.begin());
  }
}

// move everything on the level
void 
LMove() 
{
  for(int i=0; i<bl.size(); i++) {
    if(bl[i]->Mv()) {
      delete bl[i];
      bl.erase(bl.begin()+i);
      i--;
    }
  }
}

// draw everything on the same screen as the player ship
void 
LDraw() 
{
  st[sh->my*dim + sh->mx]->Dr();
  
  for(int i=0; i<bl.size(); i++) 
    if (bl[i]->mx == sh->mx && bl[i]->my == sh->my)
      bl[i]->Dr();
}

// draw player map
void
Map()
{
  int i, j, x=W-dim*8-10, y=12;
  for(j=0; j<dim; j++) {
    for(i=0; i<dim; i++) {
      rect(fb, i*8+x, j*8+y, i*8+8+x, j*8+8+y, 2);
    }
    rect(fb, sh->mx*8+x, sh->my*8+y, sh->mx*8+8+x, sh->my*8+8+y, 11);
  }
}

// game timer interrupt
void 
Tmr(...)
{
  tmr++;
}

END_OF_FUNCTION(Tmr);

// explosion generation - modified from the allegro demo game
void 
GenExpl()
{
  BM b;
  unsigned char *p;
  int i, j, x, y, xx, yy, c;

  Pt h[HS];

  for (i=0; i<HS; i++) {
    h[i].x = h[i].y = (ES/2)<<16;

    x = (rand()&0xff)<<16; //angle
    y = (rand()&0x7fff)*2; //power
    h[i].dx = fmul(fcos(x), y);
    h[i].dy = fmul(fsin(x), y);
  }

  for (i=0; i<EF; i++) {
    ex[i] = create_bitmap(ES, ES);
    b = ex[i];
    clear(b);

    c = ((i<16) ? i*4 : (80-i)) >> 2;

    for (j=0; j<HS; j++) {
      for (x=-6; x<=6; x++) {
	for (y=-6; y<=6; y++) {
	  xx = (h[j].x>>16) + x;
	  yy = (h[j].y>>16) + y;
	  if (xx>0 && yy>0 && xx<ES && yy<ES) {
	    p = b->line[yy] + xx;
	    *p += (c >> ((ABS(x)+ABS(y))/3));
	    if (*p > 63)
	      *p = 63;
	  }
	}
      }
      h[j].x += h[j].dx;
      h[j].y += h[j].dy;
    }
    
    for (x=0; x<ES; x++) {
      for (y=0; y<ES; y++) {
	j = b->line[y][x];
	if (j < 8)
	  b->line[y][x] = 0;
	else
	  b->line[y][x] = 224+j/4;
      }
    }
  }

}

// main starts here
int main()
{
  // allegro init
  allegro_init();
  install_keyboard();
  install_timer();
  text_mode(-1);
  set_gfx_mode(GFX_AUTODETECT, W, H, 0, 0);
  fb = create_bitmap(W, H); 

  dt = load_datafile("strike.dat");
  set_pallete((PALETTE)dt[7].dat);

  // interrupts
  LOCK_VARIABLE(tmr);
  LOCK_FUNCTION(Tmr);
  install_int_ex(Tmr, BPS_TO_TIMER(100));

  // other init
  srand(time(NULL));
  
  bg = create_bitmap(32,32);
  clear(bg);
  draw_sprite(         bg, (BM)dt[3].dat, 1,  1);
  draw_sprite_v_flip(  bg, (BM)dt[3].dat, 1,  16);
  draw_sprite_h_flip(  bg, (BM)dt[3].dat, 16, 1);
  draw_sprite_vh_flip( bg, (BM)dt[3].dat, 16, 16);
  
  GenExpl();

  // main loop
  do {
    sc=0;
    li=4;

    // l: current wave number (level)
    // w: timer for displaying wave message
    // done: gameover flag
    int l=1, w=2000, done=0;

    LCreate(l);

    while (!key[KEY_ESC] && !done) {
      while (tmr>0 && died!=1) {
        if (key[KEY_P]) {
          while (key[KEY_P]);
          while (!key[KEY_P]);
          while (key[KEY_P]) tmr=0;
        }
	LMove();
	tmr--;
	if (tmr>10) tmr=0;
	if (fin>0) {
	  fin--;
	  if(!fin) {
	    LDestroy();

	    clear_keybuf();
	    while(!keypressed());
	    l++;
	    LCreate(l);
	    tmr=0;
	    w=400;
	  }
	}
      }

      if (died!=1) {
        // player still alive so draw everything
	clear(fb);
	LDraw();
	Map();
	textprintf(fb, font, 0, 0, 6, 
		   " %06i0     lives:%i     wave:%i     energy:%04i      generators:%i     sector:%i%i",
                   sc, li, wv, died ? 0 : sh->hp, gn, sh->my, sh->mx);
	
	if(fin>0 && fin <150) 
	  TC(fb, "Wave Complete!", H/2, 11);

	if(sh->mx || sh->my) w=0;

	if(w) {
	  w--;
	  if (wv==1) {
	    TC(fb, "Star Strike 2000", H/2-80, 11);
	    TC(fb, "by Kevin Dixon", H/2-60, 11);
	    TC(fb, "for Allegro SizeHack 2000", H/2-40, 11);
	  }
	  textprintf_centre(fb, font, W/2, H/2, 6, "Wave %i", wv);
	}
	
	blit(fb, screen, 0, 0, 0, 0, W, H);
      }
      else { // player died

        // remove any remaining drones
	for(int i=0; i<bl.size(); i++) {
	  if(dynamic_cast<Drone*>(bl[i])) {
	    delete bl[i];
	    bl.erase(bl.begin()+i);
	    i--;
	  }
	}

	SCreate();
	li--;
	if (li<0) done=1;
      }      
    };

    LDestroy();
    TC(screen, "Game Over!", H/2+50, 2);
    clear_keybuf();
    while(!keypressed());

  } while((readkey() >>8) != KEY_ESC);

  remove_int(Tmr);
  destroy_bitmap(fb);

}

extern "C"
{
  END_OF_MAIN();
}
